static char *recvmsg_id = "@(#) $Id: recvmsg.c,v 1.13 2000/03/01 05:09:42 kouno Exp $";

/* 
 *  Project: IPMSG
 *  Author & Copyright:  KOHNO TAKASHI(͖엲) <kouno@catnet.ne.jp>
 *  Mail: kouno@catnet.ne.jp
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <netdb.h>
#include <signal.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#if defined(DEBUG)
#define line	fprintf(stderr, "Line %d\n", __LINE__);fflush(stderr);fflush(stdout)
#define d_printf printf
#else
#define debug
#define d_printf dummy_function
#endif

#define MAILMSG_VERSION "0.54"

#define IPMSG_VERSION		0x001
#define IPMSG_PORT			0x979
/* #define IPMSG_PORT 2000 */
#define IPMSG_REVERSEICON	0x100

/*  command  */
#define IPMSG_NOOPERATION 0x00000000UL

#define IPMSG_BR_ENTRY 0x00000001UL
#define IPMSG_BR_EXIT 0x00000002UL
#define IPMSG_ANSENTRY 0x00000003UL
#define IPMSG_BR_ABSENCE 0x00000004UL

#define IPMSG_BR_ISGETLIST 0x00000010UL
#define IPMSG_OKGETLIST 0x00000011UL
#define IPMSG_GETLIST 0x00000012UL
#define IPMSG_ANSLIST 0x00000013UL
#define IPMSG_BR_ISGETLIST2 0x00000018UL

#define IPMSG_SENDMSG 0x00000020UL
#define IPMSG_RECVMSG 0x00000021UL
#define IPMSG_READMSG 0x00000030UL
#define IPMSG_DELMSG 0x00000031UL

#define IPMSG_GETINFO 0x00000040UL
#define IPMSG_SENDINFO 0x00000041UL

#define IPMSG_GETABSENCEINFO 0x00000050UL
#define IPMSG_SENDABSENCEINFO 0x00000051UL

/*  option for all command  */
#define IPMSG_ABSENCEOPT 0x00000100UL
#define IPMSG_SERVEROPT 0x00000200UL
#define IPMSG_DIALUPOPT 0x00010000UL

/*  option for send command  */
#define IPMSG_SENDCHECKOPT 0x00000100UL
#define IPMSG_SECRETOPT 0x00000200UL
#define IPMSG_BROADCASTOPT 0x00000400UL
#define IPMSG_MULTICASTOPT 0x00000800UL
#define IPMSG_NOPOPUPOPT 0x00001000UL
#define IPMSG_AUTORETOPT 0x00002000UL
#define IPMSG_RETRYOPT 0x00004000UL
#define IPMSG_PASSWORDOPT 0x00008000UL
#define IPMSG_NOLOGOPT 0x00020000UL
#define IPMSG_NEWMUTIOPT 0x00040000UL
#define IPMSG_NOADDLISTOPT 0x00080000UL

/*  end of IP Messenger Communication Protocol version 1.0 define  */

#define UDP_BUF 8192
#define MAXBUFLEN 0x2ff
#define BROADCAST "DEFAULT"
#define TIPMSGDIR_PREFIX "tipmsg"
#define CONFIG_FILE TIPMSGDIR_PREFIX ".conf"
#define RECVMSGBC "recvmsgbc"
#define RECVMSGBC_BWC ".recvmsgbc"
#define ENV_HOME "HOME"
#define ENV_OSTYPE "OSTYPE"
#define ENV_HOSTTYPE "HOSTTYPE"

#define DEFAULT_USERLIST "userlist"
#define DEFAULT_SAVEFILE "save"
#define DEFAULT_LOGFILE "recvmsg.log"
#define DEFAULT_RECENTFILE "recent_recvfrom"
#define DEFAULT_ABSENCEFILE "absence.txt"
#define ABSENCE_MESSAGE "s݂Ⴂ܂I"
#define RECVMSG_PRT "%s@%s 烁bZ[W̎MmF󂯎܂\n"
#define READMSG_PRT "%s@%s 烁bZ[W̊JmF󂯂Ƃ܂\n"
#define EXIT_PRT "Exit (%s@%s[%s])\n"
#define ENTRY_PRT "Entry (%s@%s%s%s[%s])\n"

#define MAXBCNUM 0x100
#define MAXBCLEN 16
#define MAXSTPACKET 100
#define MAX_RETRY 1
#define RETRY_WAIT 2
#define SAVE_FILE_NUM_MAX "3"

#define GET_MODE(command)	(command & 0x000000ff)
#define GET_OPT(command)	(command & 0x0000ff00)

enum {MSG_NORMAL = 0, MSG_QUIT=1, MSG_ABSENCE};

typedef struct {
    int   format_ver;
    int   packet_no;
    char  msg_user[MAXBUFLEN];
    char  msg_host[MAXBUFLEN];
    int   command;
    int   group_offset;
    char  extend[UDP_BUF];
} MSG;

typedef struct USERLIST_TMP {
    struct USERLIST_TMP* next;
    struct USERLIST_TMP* prev;
	struct sockaddr_in	saddr;

    MSG msg; /* Must last location */
} USERLIST;

typedef struct {
    USERLIST* first;
    USERLIST* last;
} USERLIST_MNG;

int sender(int sd, char *buf, long addr, int command);
int recver(int sd);
long hosttoaddr(char *target);
int decom_msg(MSG *msg, char *buf);
void interrupt_exit();
void print_usage(void);
int check_message(int sd, MSG* msg, struct sockaddr_in* psaddr);
int store_message(MSG *msg, struct sockaddr_in* psaddr);
int print_message(FILE* fp, MSG* msg, struct sockaddr_in* psaddr);
int send_file(int sd, long addr, char* fname, int command);
int analyze_option(int* argc_p, char*** argv_p);
int load_config(char* fname);

int append_userlist(USERLIST_MNG* ulm, USERLIST* ul);
int delete_userlist(USERLIST_MNG* ulm, USERLIST* ul);
void destroy_userlist(USERLIST_MNG* ulm);
USERLIST* search_userlist(USERLIST* uls, USERLIST* ul, int flag);
USERLIST* search_userlist2(USERLIST* uls, long addr, int flag);
int flush_userlist(char* path, USERLIST_MNG* ulm);
int recent_recvfrom_write(struct sockaddr_in saddr, MSG* pmsg);

void dummy_function(char *dummy,...);

/* Lϐ  */
int g_sd;
int g_ipmsg_port = IPMSG_PORT;
int g_broadcast_num = 0;
char g_broadcast[MAXBCNUM][MAXBCLEN];
int g_get_packet_num = 0;
int g_storage_packet[MAXSTPACKET];
int g_nflag = 1;
int g_timer_exit = 0;
char g_save_file_master[PATH_MAX] = "";
char g_save_file_number[PATH_MAX] = "";
char g_log_file[PATH_MAX] = "";
char g_userlist_file[PATH_MAX] = "";
int g_exit = 0;
char g_name[MAXBCNUM] = "";
char g_group[MAXBCNUM] = "";
USERLIST_MNG g_userlist_mng = {NULL, NULL};
char g_recent_recvfrom_file[PATH_MAX] = "";
int g_absence = 0;
int g_readmsg = IPMSG_READMSG;
char g_absence_msg_file[PATH_MAX] = "";
char g_vermsg[PATH_MAX] = "";
char g_tipmsgdir[PATH_MAX] = "";
int g_nostdout = 0;
/* Lϐ ܂ */

int main(int argc, char *argv[])
{
	int	sd;
    char* ptr;
    int lp;
    char bcfile[PATH_MAX];
    int flag=1;
    int ret = 0;
	char *ostype, *hosttype;
	struct sockaddr_in	saddr;

	hosttype=getenv(ENV_HOSTTYPE);
	ostype=getenv(ENV_OSTYPE);
	sprintf(g_vermsg, "recvmsg %s for %s %s\n\n\nBuild %s", 
			MAILMSG_VERSION,  hosttype ? hosttype : "Unknown Host", 
			ostype ? ostype : "Unknown OS", recvmsg_id);
    memset(g_storage_packet, 0, MAXSTPACKET);
		
	/* NIvV̉ */
	if (!analyze_option(&argc, &argv)){
		return 0;
	}
	if (!g_absence_msg_file[0]){
		sprintf(g_absence_msg_file, "%s/%s", g_tipmsgdir, DEFAULT_ABSENCEFILE);
	}
	
	/* ݒt@C̊mF */
    mkdir(g_tipmsgdir, 0755);
    if (argc>1){
        ptr = argv[1];
    } else {
		ptr = bcfile;
		sprintf(bcfile, "%s/%s", g_tipmsgdir, CONFIG_FILE);
        if (access(ptr, F_OK) != 0){
            fprintf(stderr, "Waring: Not found [%s].\n"
					"Try Using old style file -> ", ptr);
			sprintf(bcfile, "%s/%s", g_tipmsgdir, RECVMSGBC);
			fprintf(stderr, "%s\n", ptr);
		}
		if (access(ptr, F_OK) != 0){
            fprintf(stderr, "Waring: Not found [%s].\n"
					"Try Using old^2 style file -> ", ptr);
            sprintf(bcfile, "%s/%s", getenv(ENV_HOME), RECVMSGBC_BWC);
			fprintf(stderr, "%s\n", ptr);
        }
    }
    load_config(ptr);

	/* [U[̐ݒ */
	if (!g_name[0]){
        strcpy(g_name, getpwuid(getuid())->pw_name);
    }

	/* \Pbg̏ */
	if ((sd = socket(AF_INET, SOCK_DGRAM, 0)) ==-1){
		perror("Can't create socket\n");
	}
	
    flag = 1;
    if (setsockopt(sd, SOL_SOCKET, SO_BROADCAST, (char*)&flag, sizeof(int)) == -1) {
        perror("setsockopt fail(SO_BROADCAST)");
    }
    flag = 1;
    if (setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, (char*)&flag, sizeof(int)) == -1) {
        perror("setsockopt fail(SO_KEEPALIVE)");
    }
	/* sockaddr_in\̂̐ݒ */
	memset((char *)&saddr, 0, sizeof(saddr));
	saddr.sin_family		= AF_INET;
	saddr.sin_port			= htons(g_ipmsg_port);
	saddr.sin_addr.s_addr	= htonl(INADDR_ANY);
	/* \PbgoCh */
	if (bind(sd, (struct sockaddr *)&saddr, sizeof(saddr)) == -1){
		fprintf(stderr, "Bind error %d port\n", g_ipmsg_port);
		perror(NULL);
		goto ERR2;
	}
    /* SIGNAL */
    g_sd = sd;
    signal(SIGTERM, interrupt_exit);
    signal(SIGINT,  interrupt_exit);
    /* Timer */
    if (g_timer_exit){
        struct itimerval value;
        signal(SIGALRM, interrupt_exit);

        value.it_interval.tv_sec = g_timer_exit;
        value.it_interval.tv_usec = 0;
        value.it_value.tv_sec = g_timer_exit;
        value.it_value.tv_usec = 0;
        setitimer(ITIMER_REAL, &value, NULL);
    }
	/* M҂ */
	if (!recver(sd)){
		ret = 1;
	}

  ERR2:	/* I */
    if (g_userlist_file[0]){
        destroy_userlist(&g_userlist_mng);
    }

    for (lp=0; lp<g_broadcast_num; lp++){
        sender(sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
    }
	close( sd );
	return ret;
}

/* ̉ */
int analyze_option(int* argc_p, char*** argv_p)
{
    char* ptr;
	int lp;
	char dir_prospect[PATH_MAX];
	struct stat statbuf;

	/* -p (p|[g)͍ŏɏ */
	sprintf(g_tipmsgdir, "%s/%s", getenv(ENV_HOME), TIPMSGDIR_PREFIX);
	for(lp=1 ; lp<(*argc_p) && (*argv_p)[lp][0] == '-'; lp++){
		if (tolower((*argv_p)[lp][1]) != 'p') continue;
			g_ipmsg_port=strtol(&(*argv_p)[lp][2], &ptr, 0);
			fprintf(stderr, "PORT %d(%s)\n", g_ipmsg_port, &(*argv_p)[lp][0]);
			if (g_ipmsg_port==0)
				g_ipmsg_port=IPMSG_PORT;
			sprintf(dir_prospect, "%s/%s.%d", 
					getenv(ENV_HOME), TIPMSGDIR_PREFIX, g_ipmsg_port);
			if (stat(dir_prospect, &statbuf) == 0 && 
				S_ISDIR(statbuf.st_mode) && (statbuf.st_mode & S_IREAD)){
				strcpy(g_tipmsgdir, dir_prospect);
				printf("DIR: %s\n", g_tipmsgdir);
			}
			break;
	}
	
	/* ̉ - nCt1ɂ1̃o[W*/
	for( ;(*argc_p) > 1 && (*argv_p)[1][0] == '-';(*argv_p)++, (*argc_p)-- ){
		switch( (*argv_p)[1][1] ){
		case '?':
		case 'h':
			/* wv */
			print_usage();
			return 0;
		case 'N':
			fclose(stdout);
		case 'n':
			/* ENTRY, EXIT ̕\} */
			g_nflag = 0;
			break;
		case 'e':
		case 'E':
			/* ^C}I */
			g_timer_exit = strtol(&(*argv_p)[1][2], &ptr, 0);
			fprintf(stderr, "TIMER EXIT %d(%s)\n", g_timer_exit, &(*argv_p)[1][0]);
			break;
		case 'f':
		case 'F':
			/* ʃt@Cւ̃MO */
			if ((*argv_p)[1][2] != '\0') {
				strcpy(g_save_file_master, &(*argv_p)[1][2]);
			} else {
				sprintf(g_save_file_master, "%s/%s", 
						g_tipmsgdir, DEFAULT_SAVEFILE);
				mkdir(g_save_file_master, 0700);
			}
			d_printf("Enable file saving(%s)\n", g_save_file_master);
			break;
		case 'l':
		case 'L':
			/* 1t@Cւ̃MO */
			if ((*argv_p)[1][2] != '\0') {
				strcpy(g_log_file, &(*argv_p)[1][2]);
			} else {
				sprintf(g_log_file, "%s/%s", g_tipmsgdir, DEFAULT_LOGFILE);
			}
			d_printf("Enable message loging(%s)\n", g_log_file);
			break;
		case 'u':
		case 'U':
			/* [U[Xg̍쐬Ap */
			if ((*argv_p)[1][2] != '\0') {
				strcpy(g_userlist_file, &(*argv_p)[1][2]);
			} else {
				sprintf(g_userlist_file, "%s/%s", 
						g_tipmsgdir, DEFAULT_USERLIST);
			}
			d_printf("Enable Userlist(%s)\n", g_userlist_file);
			break;
		case 'r':
		case 'R':
			/* ŐVM҂̋L^ */
			if ((*argv_p)[1][2] != '\0') {
				strcpy(g_recent_recvfrom_file, &(*argv_p)[1][2]);
			} else {
				sprintf(g_recent_recvfrom_file, "%s/%s", 
						g_tipmsgdir, DEFAULT_RECENTFILE);
			}
			d_printf("Enable Recent recvfrom.(%s)\n", g_recent_recvfrom_file);
			break;
		case 's':
		case 'S':
			/* pւ̕ԓ */
			g_readmsg = 0;
			break;
		case 'a':
		case 'A':
			/* sݐݒ */
			g_absence = IPMSG_ABSENCEOPT;
			if ((*argv_p)[1][2] != '\0') {
				strcpy(g_absence_msg_file, &(*argv_p)[1][2]);
			}
			d_printf("Enable absence file.(%s)\n", g_absence_msg_file);
			break;
		default:
			break;
		}
    }
	return 1;
}

int load_config(char* fname)
{
	FILE* fp;
	char buf[UDP_BUF];
	char* p;

    if ((fp = fopen(fname, "r")) != NULL){
        while(fgets(buf, UDP_BUF, fp)){
            d_printf("Read %s.\n", buf);
            if (buf[0] == '#' || buf[0] == '\n'){
                /* COMMENT */
                continue;
            }
            for (p = buf; *p != '\0'; p++){
                /* Translate buffer last */
                if (*p == '\n' || *p == '\r'){
                    *p = '\0';
                    break;
                }
            }
            
            /* Selected buffer contents */
            if (buf[0] == '!' || toupper(buf[0]) == 'N'){
                /* Username detected */
                for (p=buf + 1; *p != '\0'; p++){
                    if (*p != '\t' && *p != ' '){
                        strcpy(g_name, p);
                        d_printf("Username detected(%s):%s\n", g_name,p);
                        break;
                    }
                }
            } else if (buf[0] == '%' || toupper(buf[0]) == 'G'){
                /* Groupname detected */
                for (p=buf + 1; *p;p++){
                    if (*p != '\t' && *p != ' '){
                        strcpy(g_group, p);
                        d_printf("Groupname detected(%s)\n", g_group);
                        break;
                    }
                }
            } else if (buf[0] < '0' || buf[0] > '9'){
                /* Invalid buffer */
                d_printf("Invalid buffer contents : %s\n", buf);
            } else {
                /* Broadcast address */
                if (g_broadcast_num > MAXBCNUM){
                    d_printf("Too many broadcast address.\n");
                    break;
                }
                strcpy(g_broadcast[g_broadcast_num++], buf);
                d_printf("%d: Broadcast address detected.(%s)\n", g_broadcast_num, buf);
            }
        }
        fclose(fp);
    } else {
        fprintf(stderr, "Can't open %s file : ", fname);
        perror(NULL);
        g_broadcast_num = 1;
        strcpy(g_broadcast[0], BROADCAST);
    }
    return 1;
}

/* Sender */
int sender( int sd, char *buf, long addr, int command)
{
	char full_buf[UDP_BUF];
    char hostname[MAXBUFLEN];
	static long packet_no = 0;
	struct sockaddr_in	saddr;
    char* p;

    if (!packet_no){
        packet_no = time(&packet_no) + ((long) getpid() << 16);
    } else {
        packet_no++;
    }

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family		= AF_INET;
	saddr.sin_port			= htons(g_ipmsg_port);
	saddr.sin_addr.s_addr	= addr;

	if (gethostname(hostname, MAXBUFLEN) == -1){
		strcpy(hostname,"Unknow host");
	}
    if ((p = strchr(hostname, '.')) != NULL){
        *p = '\0';
    }
    
	/* Make Message */
	if (buf){
		if (strlen(buf)>UDP_BUF-100){
			buf[UDP_BUF-100]='\0';
		}
	}
	
	sprintf(full_buf, "%d:%ld:%s:%s:%d:%s", IPMSG_VERSION, packet_no, getpwuid(getuid())->pw_name, hostname, command, buf ? buf : "");
    
	/* Send */
	if (sendto(sd, full_buf, strlen(full_buf) +1, 0, (struct sockaddr *)&saddr, sizeof(saddr)) == -1){
		perror("sendto err");
		return	0;
	}
    
	return 1;
}

int entry_sender( int sd, char* name, char* group, long addr, int command)
{
	char buf[UDP_BUF];
    char hostname[MAXBUFLEN];
	static int packet_no;
	struct sockaddr_in	saddr;
    char* p;
    int len;

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family		= AF_INET;
	saddr.sin_port			= htons(g_ipmsg_port);
	saddr.sin_addr.s_addr	= addr;

	if (gethostname(hostname, MAXBUFLEN) == -1){
		strcpy(hostname,"Unknow host");
	}
    if ((p = strchr(hostname, '.')) != NULL){
        *p = '\0';
    }

	/* Make Message */
	memset(buf, 0, UDP_BUF);
	sprintf(buf, "%d:%d:%s:%s:%d:%s%c%s%c", IPMSG_VERSION, packet_no++, getpwuid(getuid())->pw_name, hostname, command|g_absence, name ? name : "", '\0', group ? group : "", '\0');

	/* Send */
    len = strlen(buf);
    len += strlen(buf + len + 1) + 2;
	if (sendto(sd, buf, len, 0, (struct sockaddr *)&saddr, sizeof(saddr)) == -1){
		perror("[ENTRY] sendto err");
		return	0;
	}
	return 1;
}

/* zXg -> AhX ϊ */
long hosttoaddr(char *target)
{
	long addr;
	struct hostent *host;

    if (strcmp(target, BROADCAST) == 0){
        return ~0;
    }

	if ((addr = inet_addr(target)) == -1){
		if ((host = (struct hostent *)gethostbyname(target)) != NULL){
			addr = *(long *)host->h_addr;
		} else {
            fprintf(stderr, "invalid hostname %s\n", target);
			perror(NULL);
			return 0;
		}
	}
    
	return addr;
}

/* Receive Message function */
int recver(int sd)
{
	char buf[UDP_BUF];
	struct sockaddr_in	saddr;
	MSG msg;
	int flag=1;
	int	size = sizeof(struct sockaddr_in);
	int lp;
    fd_set rfds;
    struct timeval tv;
    int retval;
    USERLIST* ul;
    int ff=0; /* Flush userlist flag */

	/* 擪 USER LIST ̗̈m */
	if (!(ul = (USERLIST*)calloc(1, sizeof(USERLIST)))){
		perror("Message memory allocate error!\n");
        goto ERR0;
	}
    msg.extend[0] = '\0';

	/* Gg[ʒm */
    for (lp=0; lp<g_broadcast_num; lp++){
        if (!entry_sender(sd, g_name, g_group, hosttoaddr(g_broadcast[lp]), IPMSG_BR_ENTRY)){
            goto ERR2;
        }
    }

	while(flag){
        if (g_exit){
            flag = 0;
        }
        FD_ZERO(&rfds);
        FD_SET(sd, &rfds);
        tv.tv_sec = 1;
        tv.tv_usec = 0;
        
        /* tJn */
        retval = select(sd+1, &rfds, NULL, NULL, &tv);
        if (retval == -1){
            perror("select fail");
            flush_userlist(g_userlist_file, &g_userlist_mng);
            goto ERR2;
        } else if (!retval){
            if (ff){
				/* USER LIST ̏ */
                flush_userlist(g_userlist_file, &g_userlist_mng);
                ff=0;
            }
            continue;
        }
        
        /* f[^擾 */
		if ((retval = recvfrom(sd, buf, UDP_BUF -1, 0, (struct sockaddr *)&saddr, &size)) == -1){
            perror("recv error");
            goto ERR2;
		}
		buf[retval]='\0';

		if (!decom_msg(&msg, buf)){
			continue;
		}
        /* Mobt@̃[NA(e͊ msg \̂ɑ݂) */
        memset(buf, 0, retval);

		/* R}h̔ */
		d_printf("Command %08x\n", msg.command);
		switch(GET_MODE(msg.command)){
        case IPMSG_BR_ENTRY:
            if (!entry_sender(sd, g_name, g_group, saddr.sin_addr.s_addr, 
							  IPMSG_ANSENTRY)){
                goto ERR2;
            }
            /* ʂ蔲` */
        case IPMSG_ANSENTRY:
        	if (g_nflag){
	            fprintf(stdout, ENTRY_PRT, 
						msg.extend[0] ? msg.extend : msg.msg_user, 
						msg.extend + msg.group_offset, 
						msg.extend[msg.group_offset] ? "/" : "", 
						msg.msg_host, inet_ntoa(saddr.sin_addr));
	        }
            ul->msg = msg;
            ul->saddr = saddr;
            if (g_userlist_file[0]){
                append_userlist(&g_userlist_mng, ul);
            }
            ff=1;
            break;
        case IPMSG_BR_EXIT:
        	if (g_nflag){
	            fprintf(stdout, EXIT_PRT, msg.msg_user, msg.msg_host, 
						inet_ntoa(saddr.sin_addr));
	        }
            ul->msg = msg;
            ul->saddr = saddr;
            if (g_userlist_file[0]){
                delete_userlist(&g_userlist_mng, ul);
            }
            ff=1;
            break;
        case IPMSG_RECVMSG:
            fprintf(stdout, RECVMSG_PRT, msg.msg_user, msg.msg_host);
            break;
        case IPMSG_READMSG:
            fprintf(stdout, READMSG_PRT, msg.msg_user, msg.msg_host);
            break;
		case IPMSG_GETINFO:
			sender(sd, g_vermsg, saddr.sin_addr.s_addr, 
				   IPMSG_AUTORETOPT|IPMSG_SENDINFO);
			break;
		case IPMSG_GETABSENCEINFO:
			if (g_absence){
				send_file(sd, saddr.sin_addr.s_addr, g_absence_msg_file, 
						  IPMSG_AUTORETOPT|IPMSG_SENDABSENCEINFO);
			} else {
				sender(sd, ABSENCE_MESSAGE, saddr.sin_addr.s_addr, 
					   IPMSG_AUTORETOPT|IPMSG_SENDABSENCEINFO);
			}
			
			break;
        case IPMSG_SENDMSG:
			switch (check_message(sd, &msg, &saddr)){
			case MSG_NORMAL:
				store_message(&msg, &saddr);
				break;
			case MSG_QUIT:
				goto ERR2;
			default:
				break;
			}
            break;
		default:
			d_printf("Recive unknow command: 0x%08x\n", msg.command);
		}
        fflush(stdout);
	}
	return 1;

 ERR2:
    free(ul);
 ERR0:
    return 0;
}

/* t@C̑M */
int send_file(int sd, long addr, char* fname, int command)
{
	struct stat sbuf;
	int len;
	char buf[UDP_BUF];
	int fd;

	d_printf("START send_file(%s)\n", fname);
	if (stat(fname, &sbuf) != 0){
		if (errno != ENOENT){
			fprintf(stderr, "Can't get file status(%s) : %s\n", 
					fname, strerror(errno));
		}
		
		return 0;
	}
	if (!sbuf.st_size){
		d_printf("File size is Zero(%s).\n", fname);
		return 1;
	}
	if ((fd = open(fname, O_RDONLY)) <0){
		fprintf(stderr, "Can't open file(%s) : %s\n", fname, strerror(errno));
		return 0;
	}
	len = (sbuf.st_size < UDP_BUF) ? sbuf.st_size : UDP_BUF;
	if (read(fd, buf, len)<0){
		fprintf(stderr, "Can't open file(%s) : %s\n", fname, strerror(errno));
		return 0;
	}
	buf[len]='\0';
	return sender(sd, buf, addr, command);
}

/* ŏIM҂̏ */
int recent_recvfrom_write(struct sockaddr_in saddr, MSG* pmsg)
{
    char recent_recvfrom[PATH_MAX];
	USERLIST* ul;
	FILE* fp;
	
	if ((ul = search_userlist2(g_userlist_mng.first, saddr.sin_addr.s_addr, 0)) != NULL){
		sprintf(recent_recvfrom, "%s@%s%s%s[%s](%s:%s)", 
				ul->msg.extend[0] ? ul->msg.extend : ul->msg.msg_user, 
				ul->msg.extend + ul->msg.group_offset, 
				ul->msg.extend[ul->msg.group_offset] ? "/" : "", 
				ul->msg.msg_host, inet_ntoa(ul->saddr.sin_addr), 
				ul->msg.msg_user, ul->msg.msg_host);
	} else {
		d_printf("Can't find message user in userlist.\n");
		fprintf(stdout, "%s@%s%s%s[%s]\n", 
				pmsg->extend[0] ? pmsg->extend : pmsg->msg_user, 
				pmsg->extend + pmsg->group_offset, 
				pmsg->extend[pmsg->group_offset] ? "/" : "", 
				pmsg->msg_host, inet_ntoa(saddr.sin_addr));
	}
	if ((fp = fopen(g_recent_recvfrom_file, "w")) != NULL){
		fprintf(fp, recent_recvfrom);
		fclose(fp);
		d_printf("Set recent recvfrom.(%s)\n", recent_recvfrom);
	} else {
		fprintf(stderr, "Can't open recent recvfrom file.(%s)\n", g_recent_recvfrom_file);
	}
	return 1;
}

/* Mobt@̃bZ[W\̂ւ̓WJ */
int decom_msg( MSG *msg, char *buf )
{
	char *str;
	
	/* Get format version */
	if ((str=strchr( buf,':')) == NULL){
		fprintf(stderr, "Can't find format version\n" );
		return 0;
	}
	str[0]='\0';
	if ((msg->format_ver=atoi(buf)) != IPMSG_VERSION){
		return 0;
	}
	buf=str+1;
	
	/* Get Packet No */
	if ((str=strchr(buf, ':')) == NULL){
		fprintf(stderr, "Can't find Packet\n");
		return 0;
	}
	str[0]='\0';
	msg->packet_no=atoi( buf ) ;
	buf=str+1;
	
	/* Get Sender Name */
	if ((str=strchr(buf, ':')) == NULL){
		fprintf(stderr, "Can't find Sender Name\n" );
		return 0;
	}
	str[0]='\0';
	strcpy( msg->msg_user, buf );
	buf=str+1;
	
	/* Get Sendhost Name */
	if ((str=strchr(buf,':')) == NULL){
		fprintf(stderr, "Can't find Sendhost Name\n" );
		return 0;
	}
	str[0]='\0';
	strcpy( msg->msg_host, buf );
	buf=str+1;
	
	/* Get command */
	if ((str=strchr(buf,':')) == NULL){
		fprintf(stderr, "Can't find command\n" );
		return 0;
	}
	str[0]='\0';
	msg->command=atoi(buf) ;
	buf=str+1;
	
	/* Get Extension */
	sprintf(msg->extend, "%s%c%s", buf, '\0', buf + strlen(buf)+1);
    msg->group_offset = strlen(msg->extend)+1;
	
	return 1;
}

void interrupt_exit()
{
    int lp;
    
    signal(SIGINT, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    for (lp=0; lp<g_broadcast_num; lp++){
        sender(g_sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(g_sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
        sender(g_sd, NULL, hosttoaddr(g_broadcast[lp]), IPMSG_BR_EXIT);
    }
    d_printf("SIGNAL INTERRUCT.\n");
    g_exit = 1;
    signal(SIGTERM, SIG_DFL);
    signal(SIGINT, SIG_DFL);
}

/* g\ */
void print_usage()
{
    printf("resvmsg ver %s - IPMSG recever tool\n", MAILMSG_VERSION);
    printf("Usage: recvmsg -[npefh] [Broadcast file]\n");
    printf("        -n : Contain [ENTRY]and[EXIT] message.\n");
    printf("        -p : Designate port number.\n");
    printf("        -e : Designate exit timer.\n");
    printf("        -f : Designate file save.\n");
    printf("        -h : Print this message.\n");
    printf("        -u : Using userlist.\n");
    printf("Ex. recvmsg -n -p2000 -e5 -f/var/log/recvmsg -u/tmp/userlist ~/.recvmsgbc\n");
}

/* bZ[WM̏`FbN */
int check_message(int sd, MSG* msg, struct sockaddr_in* psaddr)
{
	char buf[UDP_BUF];
	int lp;

    /* M`FbN */
    if (GET_OPT(msg->command) & IPMSG_SENDCHECKOPT){
        sprintf(buf, "%d", msg->packet_no);
        if (!sender(sd, buf, psaddr->sin_addr.s_addr, IPMSG_RECVMSG)){
            return 0;
        }
    }
    if (GET_OPT(msg->command) & IPMSG_SECRETOPT){
        sprintf(buf, "%d", msg->packet_no);
		if (g_readmsg){
			if (!sender(sd, buf, psaddr->sin_addr.s_addr, IPMSG_READMSG)){
				return 0;
			}
		}
    }
	/* 擪Mɂ铮ύX */
	if (strncmp(msg->extend, "quit", 4) == 0){
		msg->extend[0]='\0';
		return MSG_QUIT;
	} else if (strncmp(msg->extend, "absence", 4) == 0){
		msg->extend[0]='\0';
		g_absence = g_absence ? 0 : IPMSG_ABSENCEOPT;
		for (lp=0; lp<g_broadcast_num; lp++){
			entry_sender(sd, g_name, g_group, hosttoaddr(g_broadcast[lp]), 
						 IPMSG_BR_ENTRY);
		}
		return MSG_ABSENCE;
	}

	/* bZ[WM҂̕ۑ */
	if (g_recent_recvfrom_file[0]){
		recent_recvfrom_write(*psaddr, msg);
	}

	/* ԓbZ[Wɂ͂ȍ~̉͂Ȃ */
	if (GET_OPT(msg->command) & IPMSG_AUTORETOPT){
		d_printf("It's IPMSG_AUTORETOPT.\n");
		return MSG_NORMAL;
	}
	/* sݎ */
	if (g_absence){
		d_printf("g_absence[%s]\n", g_absence_msg_file);
		send_file(sd, psaddr->sin_addr.s_addr, 
				  g_absence_msg_file, IPMSG_AUTORETOPT|IPMSG_SENDMSG);
	}
	return MSG_NORMAL;
}

/* Message ̕ۑ */
int store_message(MSG* msg, struct sockaddr_in* psaddr)
{
    FILE* fp;
    static int count = 0;
    char linkbuf[PATH_MAX];
    static char master[PATH_MAX]="";
    char *endp;
    char *startp;

    fp = stdout;

	/* Wo͂ւ̕ۑ*/
    fprintf(stdout, "\a\a\n" );
    print_message(stdout, msg, psaddr);

	/* Matser file ւ̕ۑƃN̍XV */
    if (g_save_file_master[0]){
        if (!master[0]){
            sprintf(master, "%s/0", g_save_file_master);
        }
        if (!count){
            if (readlink(master, linkbuf, PATH_MAX)){
                if ((startp = strrchr(linkbuf, '/')) != NULL){
                    d_printf("link to %s\n", startp);
                    count = strtol(startp+1, &endp, 10);
                }
            }
        }
        count++;
        sprintf(g_save_file_number, "%s/%0" SAVE_FILE_NUM_MAX "d", g_save_file_master, count);
        if ((fp = fopen(g_save_file_number, "w"))){
            print_message(fp, msg, psaddr);
        
            fclose(fp);
            unlink(master);
            symlink(g_save_file_number, master);
        } else {
            fprintf(stderr, "Can't open save file(%s)", g_save_file_number);
            perror(NULL);
        }
    }

	/* Ot@Cւ̕ۑ */
    if (g_log_file[0]){
        if ((fp = fopen(g_log_file, "a"))){
            print_message(fp, msg, psaddr);
        
            fclose(fp);
            unlink(master);
            symlink(g_save_file_number, master);
        } else {
            fprintf(stderr, "Can't open log file(%s)", g_save_file_number);
            perror(NULL);
        }
    }

    return 1;
}

/* bZ[W̕\ */
int print_message(FILE* fp, MSG* msg, struct sockaddr_in* psaddr)
{
	USERLIST* ul;
    char user[PATH_MAX];
    char host[PATH_MAX];
	char horizon[] = "----------------------------------------------\n";
	char horizon_end[] = "----------------Message End-------------------\n\n";
	char buf[PATH_MAX];
	int lp;

    time_t now_time;
    time( &now_time );

	if ((ul = search_userlist2(g_userlist_mng.first, psaddr->sin_addr.s_addr, 0)) != NULL){
		strcpy(user,ul->msg.extend[0] ? ul->msg.extend : ul->msg.msg_user);
		sprintf(host, "%s%s%s", ul->msg.extend + ul->msg.group_offset, 
				ul->msg.extend[ul->msg.group_offset] ? "/" : "", 
				ul->msg.msg_host);
	} else {
		strcpy(user, msg->msg_user);
		strcpy(host, msg->msg_host);
	}

    fprintf(fp, horizon);
    
/* Z^OԊuvZ buf ̕\ */
#define PM2CENTER if (sizeof(horizon) > strlen(buf)) for (lp=0;  lp < (sizeof(horizon) - strlen(buf)) / 2; lp++) fprintf(fp, " ");fprintf(fp, buf)

    if (GET_OPT(msg->command) & IPMSG_BROADCASTOPT){
		/* u[hLXg */
		sprintf(buf, "Broadcast from %s ( %s )\n", user, host);
		PM2CENTER;
    } else {				
        if (GET_OPT(msg->command) & IPMSG_MULTICASTOPT){
			/* }`LXg */
            sprintf(buf, "Multicast from %s ( %s )\n", user, host);
			PM2CENTER;
        } else {
			/* ʏ */
            sprintf(buf, "Message from %s ( %s )\n", user, host );
			PM2CENTER;
        }
    }
    sprintf(buf, "at %s", ctime(&now_time));
	PM2CENTER;
    fprintf(fp, horizon);
    fprintf(fp, "%s\n", msg->extend);
    fprintf(fp, horizon_end);

    return 1;
}

/* USER LIST ւ̒ǉ */
int append_userlist(USERLIST_MNG* ulm /* Userlist Last */, USERLIST* ul)
{
    USERLIST* ult; /* Userlist Target*/
    if ((ult = search_userlist(ulm->last, ul, 1)) != NULL){
        /* User exist */
        return 1;
    }

    d_printf("Append %s@%s\n", ul->msg.msg_user, ul->msg.msg_host);
    if (!(ult = (USERLIST*)malloc(sizeof(USERLIST)))){
        perror("Can't allocate memory.");
        return 0;
    }

    *ult = *ul;
    ult->next = NULL;
    ult->prev = ulm->last;

    if (ulm->last){
        ulm->last->next = ult;
        ulm->last = ult;
    } else {
        d_printf("Insert first\n");
        ulm->first = ulm->last = ult;
    }
    
    return 1;
}

/* USER LIST ̍폜 */
int delete_userlist(USERLIST_MNG* ulm /* Userlist First */, USERLIST* ul)
{
    USERLIST* ult; /* Userlist Target*/
    if (!(ult = search_userlist(ulm->first, ul, 0))){
        /* User exist */
        return 0;
    }

    if (ult->prev){
        ult->prev->next = ult->next;
    } else {
        ulm->first = ult->next;
    }
    if (ult->next){
        ult->next->prev = ult->prev;
    } else {
        ulm->last = ult->prev;
    }

    free(ult);
    return 1;
}

/* USER LIST ̔j */
void destroy_userlist(USERLIST_MNG* ulm)
{
    USERLIST* ult; /* Userlist Target*/

    d_printf("Destroy_userlist Start!\n");
    while((ult = ulm->first) != NULL){
        d_printf("Delete Userlist[%s]\n", ult->msg.msg_user);
        ulm->first = ult->next;
        free(ult);
    }
    d_printf("Destroy_userlist E n d!\n");
}

/* USER LIST ̌ (flag == 1 Ō̌)*/
USERLIST* search_userlist(USERLIST* uls, USERLIST* ul, int flag)
{
    if (!ul){
        return NULL;
    }

    while(uls){
        if (uls->saddr.sin_addr.s_addr == ul->saddr.sin_addr.s_addr){
            d_printf("Search user(%s)\n", inet_ntoa(ul->saddr.sin_addr));
            return uls;
        }
        if (!flag){
            uls = uls->next;
        } else {
            uls = uls->prev;
        }
    }
    return NULL;
}

/* AhXɂ USER LIST ̌ (flag == 1 Ō̌)*/
USERLIST* search_userlist2(USERLIST* uls, long addr, int flag)
{
    while(uls){
        if (uls->saddr.sin_addr.s_addr == addr){
            d_printf("Search user(%s)\n", inet_ntoa(uls->saddr.sin_addr));
            return uls;
        }
        if (!flag){
            uls = uls->next;
        } else {
            uls = uls->prev;
        }
    }
    return NULL;
}

/*  USER LIST t@C֓fo */
int flush_userlist(char* path, USERLIST_MNG* ulm)
{
    USERLIST* ul;
    FILE* fp;

    if (!path[0] || !ulm->first){
        return 0;
    }

    d_printf("flush_userlist\n");

    if (!(fp = fopen(path, "w"))){
        fprintf(stderr, "Can't open userlist file(%s):", path);
        perror(NULL);
        return 0;
    }
    
    for (ul=ulm->first; ul; ul=ul->next){
        fprintf(fp, "%s@%s%s%s[%s](%s:%s)\n", 
                ul->msg.extend[0] ? ul->msg.extend : ul->msg.msg_user, 
                ul->msg.extend + ul->msg.group_offset, 
                ul->msg.extend[ul->msg.group_offset] ? "/" : "", 
                ul->msg.msg_host, inet_ntoa(ul->saddr.sin_addr), 
                ul->msg.msg_user, ul->msg.msg_host);
    }
    fclose(fp);
    return 1;
}

void dummy_function(char *dummy,...)
{
}
